/*
** THIS SOFTWARE IS SUBJECT TO COPYRIGHT PROTECTION AND IS OFFERED ONLY
** PURSUANT TO THE 3DFX GLIDE GENERAL PUBLIC LICENSE. THERE IS NO RIGHT
** TO USE THE GLIDE TRADEMARK WITHOUT PRIOR WRITTEN PERMISSION OF 3DFX
** INTERACTIVE, INC. A COPY OF THIS LICENSE MAY BE OBTAINED FROM THE 
** DISTRIBUTOR OR BY CONTACTING 3DFX INTERACTIVE INC(info@3dfx.com). 
** THIS PROGRAM IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
** EXPRESSED OR IMPLIED. SEE THE 3DFX GLIDE GENERAL PUBLIC LICENSE FOR A
** FULL TEXT OF THE NON-WARRANTY PROVISIONS.  
** 
** USE, DUPLICATION OR DISCLOSURE BY THE GOVERNMENT IS SUBJECT TO
** RESTRICTIONS AS SET FORTH IN SUBDIVISION (C)(1)(II) OF THE RIGHTS IN
** TECHNICAL DATA AND COMPUTER SOFTWARE CLAUSE AT DFARS 252.227-7013,
** AND/OR IN SIMILAR OR SUCCESSOR CLAUSES IN THE FAR, DOD OR NASA FAR
** SUPPLEMENT. UNPUBLISHED RIGHTS RESERVED UNDER THE COPYRIGHT LAWS OF
** THE UNITED STATES.  
** 
** COPYRIGHT 3DFX INTERACTIVE, INC. 1999, ALL RIGHTS RESERVED
 *
 ** $Header: /home/cvsroot/driglide/driglide/h5/glide3/src/fifo.c,v 1.3 2003/01/15 18:49:21 anholt Exp $
 **
 */

#include <stdio.h>
#include <string.h>

#include <3dfx.h>
#include <glidesys.h>

#define FX_DLL_DEFINITION
#include <fxdll.h>
#include <glide.h>

#include "fxglide.h"
#include "fxcmd.h"


#if GDBG_INFO_ON

static const char* h3SstRegNames[] = {
  "status",
  "intrCtrl",
  "vAx",
  "vAy",
  "vBx",
  "vBy",
  "vCx",
  "vCx",
  "r",
  "g",
  "b",
  "z",
  "a",
  "s",
  "t",
  "w",
  "drdx",
  "dgdx",
  "dbdx",
  "dzdx",
  "dadx",
  "dsdx",
  "dtdx",
  "dwdx",
  "drdy",
  "dgdy",
  "dbdy",
  "dzdy",
  "dady",
  "dsdy",
  "dtdy",
  "dwdy",
  "triangleCMD",
  "reservedA",
  "FvAx",
  "FvAy",
  "FvBx",
  "FvBy",
  "FvCx",
  "FvCy",
  "Fr",
  "Fg",
  "Fb",
  "Fz",
  "Fa",
  "Fs",
  "Ft",
  "Fw",
  "Fdrdx",
  "Fdgdx",
  "Fdbdx",
  "Fdzdx",
  "Fdadx",
  "Fdsdx",
  "Fdtdx",
  "Fdwdx",
  "Fdrdy",
  "Fdgdy",
  "Fdbdy",
  "Fdzdy",
  "Fdady",
  "Fdsdy",
  "Fdtdy",
  "Fdwdy",
  "FtriangleCMD",
  "fbzColorPath",
  "fogMode",
  "alphaMode",
  "fbzMode",
  "lfbMode",
  "clipLeftRight",
  "clipBottomTop",
  "nopCMD",
  "fastfillCMD",
  "swapbufferCMD",
  "fogColor",
  "zaColor",
  "chromaKey",
  "chromaRange",
  "userIntrCmd",
  "stipple",
  "c0",
  "c1",
  "fbiPixelsIn",
  "fbiChromaFail",
  "fbiZfuncFail",
  "fbiAfuncFail",
  "fbiPixelsOut",
  "fogTable00",
  "fogTable01",
  "fogTable02",
  "fogTable03",
  "fogTable04",
  "fogTable05",
  "fogTable06",
  "fogTable07",
  "fogTable08",
  "fogTable09",
  "fogTable0a",
  "fogTable0b",
  "fogTable0c",
  "fogTable0d",
  "fogTable0e",
  "fogTable0f",
  "fogTable10",
  "fogTable11",
  "fogTable12",
  "fogTable13",
  "fogTable14",
  "fogTable15",
  "fogTable16",
  "fogTable17",
  "fogTable18",
  "fogTable19",
  "fogTable1a",
  "fogTable1b",
  "fogTable1c",
  "fogTable1d",
  "fogTable1e",
  "fogTable1f",
  "reservedB0",
  "reservedB1",
  "reservedB2",
  "colBufferAddr",
  "colBufferStride",
  "auxBufferAddr",
  "auxBufferStride",
  "reservedC",
  "clipLeftRight1",
  "clipBottomTop1",
  "reservedD0",
  "reservedD1",
  "reservedD2",
  "reservedD3",
  "reservedD4",
  "reservedD5",
  "reservedE0",
  "reservedE1",
  "reservedE2",
  "reservedE3",
  "reservedE4",
  "reservedE5",
  "reservedE6",
  "reservedE7",
  "reservedF0",
  "reservedF1",
  "reservedF2",
  "swapBufferPend",
  "leftOverlayBuf",
  "rightOverlayBuf",
  "fbiSwapHistory",
  "fbiTrianglesOut",
  "sSetupMode",
  "sVx",
  "sVy",
  "sARGB",
  "sRed",
  "sGreen",
  "sBlue",
  "sAlpha",
  "sVz",
  "sOowfbi",
  "sOow0",
  "sSow0",
  "sTow0",
  "sOow1",
  "sSow1",
  "sTow1",
  "sDrawTriCMD",
  "sBeginTriCMD",
  "reservedG0",
  "reservedG1",
  "reservedG2",
  "reservedG3",
  "reservedG4",
  "reservedG5",
  "reservedH0",
  "reservedH1",
  "reservedH2",
  "reservedH3",
  "reservedH4",
  "reservedH5",
  "reservedH6",
  "reservedH7",
  "reservedI0",
  "reservedI1",
  "reservedI2",
  "reservedI3",
  "reservedI4",
  "reservedI5",
  "reservedI6",
  "reservedI7",
  "textureMode",
  "tLOD",
  "tDetail",
  "texBaseAddr",
  "texBaseAddr1",
  "texBaseAddr2",
  "texBaseAddr38",
  "trexInit0",
  "trexInit1",
  "nccTable0-0",
  "nccTable0-1",
  "nccTable0-2",
  "nccTable0-3",
  "nccTable0-4",
  "nccTable0-5",
  "nccTable0-6",
  "nccTable0-7",
  "nccTable0-8",
  "nccTable0-9",
  "nccTable0-a",
  "nccTable0-b",
  "nccTable1-0",
  "nccTable1-1",
  "nccTable1-2",
  "nccTable1-3",
  "nccTable1-4",
  "nccTable1-5",
  "nccTable1-6",
  "nccTable1-7",
  "nccTable1-8",
  "nccTable1-9",
  "nccTable1-a",
  "nccTable1-b",
  "tChromaKeyMin",
  "tChromaKeyMax",
};

static const char * h3SstIORegNames[] = {     
  "status",
  "pciInit0",
  "sipMonitor",
  "lfbMemoryConfig",
  "miscInit0",
  "miscInit1",
  "dramInit0",
  "dramInit1",
  "agpInit",
  "tmuGbeInit",
  "vgaInit0",
  "vgaInit1",
  "dramCommand",
  "dramData",
  "reservedZ0"
  "reservedZ0"
  "pllCtrl0",
  "pllCtrl1",
  "pllCtrl2",
  "dacMode",
  "dacAddr",
  "dacData",
  "vidMaxRGBDelta",
  "vidProcCfg",
  "hwCurPatAddr",
  "hwCurLoc",
  "hwCurC0",
  "hwCurC1",
  "vidInFormat",
  "vidInStatus",
  "vidSerialParallelPort",
  "vidInXDecimDeltas",
  "vidInDecimInitErrs",
  "vidInYDecimDeltas",
  "vidPixelBufThold",
  "vidChromaMin",
  "vidChromaMax",
  "vidCurrentLine",
  "vidScreenSize",
  "vidOverlayStartCoords",
  "vidOverlayEndScreenCoord",
  "vidOverlayDudx",
  "vidOverlayDudxOffsetSrcWidth",
  "vidOverlayDvdy",
  "vgaRegister[0]",
  "vgaRegister[1]",
  "vgaRegister[2]",
  "vgaRegister[3]",
  "vgaRegister[4]",
  "vgaRegister[5]",
  "vgaRegister[6]",
  "vgaRegister[7]",
  "vgaRegister[8]",
  "vgaRegister[9]",
  "vgaRegister[a]",
  "vgaRegister[b]",
  "vidOverlayDvdyOffset",
  "vidDesktopStartAddr",
  "vidDesktopOverlayStride",
  "vidInAddr0",
  "vidInAddr1",
  "vidInAddr2",
  "vidInStride",
  "vidCurrOverlayStartAddr",
} ;


#define GEN_INDEX(a) ((((AnyPtr) a) - ((AnyPtr) gc->reg_ptr)) >> 2)

void
_grFifoWriteDebug(FxU32 addr, FxU32 val, FxU32 fifoPtr)
{
  GR_DCL_GC;
  FxU32 index = GEN_INDEX(addr);
  
  GDBG_INFO(gc->myLevel + 199, "Storing to FIFO:\n");
  GDBG_INFO(gc->myLevel + 199, "  FIFO Ptr:    0x%x : 0x%X\n", fifoPtr, gc->cmdTransportInfo.fifoRoom);  
  if (index <= 0xff) { 
    GDBG_INFO(gc->myLevel + 199, "  Reg Name:    %s\n", h3SstRegNames[index]);
    GDBG_INFO(gc->myLevel + 199, "  Reg Num:     0x%X\n", index);
  } else {
    const char* strP;
    const FxU32 offset = (addr - (FxU32)gc->reg_ptr);
    
    if (offset >= HW_TEXTURE_OFFSET) {
      strP = "Texture";
    } else if (offset >= HW_LFB_OFFSET) {
      strP = "LFB";
      index = addr;
    } else if (offset >= HW_FIFO_OFFSET) {
      strP = "Cmd FIFO";
    } else {
      strP = "Woah!";
    }
    GDBG_INFO(gc->myLevel + 199, "  %s Addr:    0x%X\n",
              strP, index);
  }
  GDBG_INFO(gc->myLevel + 199, "  Value:       0x%X 0x%X\n", (index << 2), val);
  
  GDBG_INFO(120, "        SET(0x%X, %ld(0x%X)) 0 %s (0x%X)\n",
            0x10000000UL + (FxU32)(index << 2), val, val, 
            h3SstRegNames[index & 0xFF], fifoPtr);
} /* _grFifoWriteDebug */

void
_grFifoFWriteDebug(FxU32 addr, float val, FxU32 fifoPtr)
{
  GR_DCL_GC;
  FxU32 index = GEN_INDEX(addr);

  GDBG_INFO(gc->myLevel + 200, "Storing to FIFO:\n");
  GDBG_INFO(gc->myLevel + 200, "  FIFO Ptr:    0x%x\n", fifoPtr);
  if (index <= 0xff) {
    GDBG_INFO(gc->myLevel + 200, "  Reg Name:    %s\n", h3SstRegNames[index]);
    GDBG_INFO(gc->myLevel + 200, "  Reg Num:     0x%x\n", index);
  }
  GDBG_INFO(gc->myLevel + 200, "  Value:       %4.2f\n", val);

  GDBG_INFO(120, "        SET(0x%X, %4.2f (0x%X)) 0 %s\n", 
            0x10000000UL + (FxU32)(index << 2), val, *(const FxU32*)&val, 
            h3SstRegNames[index & 0xFF]);
} /* _grFifoFWriteDebug */

extern void
_grH3FifoDump_TriHdr(const FxU32 hdrVal)
{
  GR_DCL_GC;

  /* Dump Packet Header */
  GDBG_INFO(gc->myLevel + 200, "CMD Fifo Triangle Packet (0x%X)\n", hdrVal);
  GDBG_INFO(gc->myLevel + 200, "  # Vertex: 0x%X\n", 
            (hdrVal & SSTCP_PKT3_NUMVERTEX) >> SSTCP_PKT3_NUMVERTEX_SHIFT);
  GDBG_INFO(gc->myLevel + 200, "  RGB: %s\n",
            (hdrVal & SSTCP_PKT3_PACKEDCOLOR) ? "Packed" : "Separate");
  
  GDBG_INFO(gc->myLevel + 200, "  StripMode: %s\n",
            (((hdrVal & (0x01 << 22)) == 0) ? "Strip" : "Fan"));
  
  GDBG_INFO(gc->myLevel + 200, "  Culling: %s\n",
            (((hdrVal & (0x01 << 23)) == 0) ? "Disable" : "Enable"));
  
  GDBG_INFO(gc->myLevel + 200, "  CullingSign: %s\n",
            (((hdrVal & (0x01 << 24)) == 0) ? "Positive" : "Negative"));
  
  GDBG_INFO(gc->myLevel + 200, "  PingPongSign: %s\n",
            (((hdrVal & (0x01 << 25)) == 0) ? "Normal" : "Disable"));
  
  if (GDBG_GET_DEBUGLEVEL(gc->myLevel + 200)) {
    const FxU32 temp = (hdrVal & SSTCP_PKT3_PMASK);
    int i;

    GDBG_INFO(gc->myLevel + 200, "  Params: X Y");

    for(i = 10; i <= 17; i++) {
      static const char* paramSel[] = { "RGB", "Alpha", "Z", "Wb", "W0", "ST[0]", "W1", "ST[1]" };
      
      if ((temp & (0x01UL << i)) != 0) GDBG_PRINTF("%s ", paramSel[i - 10]);
    }
    GDBG_INFO(gc->myLevel + 200, "\n");
  }
  
  {
    const FxU32 temp = (hdrVal & SSTCP_PKT3_CMD) >> SSTCP_PKT3_CMD_SHIFT;
    const char* tempStr;

    switch(temp) {
    case 0x00: tempStr = "Independent";   break;
    case 0x01: tempStr = "NewStrip";      break;
    case 0x02: tempStr = "ContinueStrip"; break;
    default:   tempStr = "Reserved";      break;
    }
    GDBG_INFO(gc->myLevel + 200, "  Command: 0x%X(%s)\n", temp, tempStr);
  }
}

void
_grErrorCallback(const char* const procName,
                 const char* const format,
                 va_list           args)
{
  static FxBool inProcP = FXFALSE;

  if (!inProcP) {
    static char errMsgBuf[1024];
    
    inProcP = FXTRUE;
    {
      extern void (*GrErrorCallback)( const char *string, FxBool fatal );

      vsprintf(errMsgBuf, format, args);
      (*GrErrorCallback)(errMsgBuf, (GETENV("FX_ERROR_FAIL") != NULL));
    }
    inProcP = FXFALSE;
  }
}

#endif /* GDBG_INFO_ON */

#if USE_PACKET_FIFO

/* Routines privately exported so that the manufacturing diags
 * and other things can do register writes etc w/o having access
 * to the glide internals etc.
 */
extern void
_grSet32(volatile FxU32* const sstAddr, const FxU32 val)
{
#define FN_NAME "_grSet32"
  GR_DCL_GC;

  GR_ASSERT(sstAddr >= gc->base_ptr);
  GR_ASSERT(sstAddr <  &SST_TMU(gc->reg_ptr, GR_TMU0)->status);

  GR_SET_EXPECTED_SIZE(sizeof(FxU32), 1);
  GR_SET_INDEX(BROADCAST_ID, gc->reg_ptr, (sstAddr - gc->reg_ptr), val);
  GR_CHECK_SIZE();
#undef FN_NAME
}

extern FxU32
_grGet32(volatile FxU32* const sstAddr)
{
#define FN_NAME "_grGet32"
  GR_BEGIN_NOFIFOCHECK_RET(FN_NAME, 88);
  GDBG_INFO_MORE(gc->myLevel, "(0x%X)\n", sstAddr);
  GR_RETURN(GR_GET(*sstAddr)); 
#undef FN_NAME
} /* _grGet32 */

#if FIFO_ASSERT_FULL
const FxU32 kFifoCheckMask = 0xFFFF;
FxU32 gFifoCheckCount = 0;
#endif

void
_grBumpNGrind() 
{
#define FN_NAME "_grCheckForBump"
  GR_BEGIN_NOFIFOCHECK(FN_NAME, 400);

  FIFO_CACHE_FLUSH(gc->cmdTransportInfo.fifoPtr);
  P6FENCE;

  /* AH FUK: the bump register can only handle a 16bit value */
  while ((gc->cmdTransportInfo.fifoPtr - gc->cmdTransportInfo.lastBump) > 0xFFFFL)
  {
    GR_CAGP_SET(bump, 0xFFFFL);
    gc->cmdTransportInfo.lastBump += 0xFFFFL ;
  }

  if (gc->cmdTransportInfo.fifoPtr - gc->cmdTransportInfo.lastBump)
    GR_CAGP_SET(bump, gc->cmdTransportInfo.fifoPtr - gc->cmdTransportInfo.lastBump);
  
  gc->cmdTransportInfo.lastBump = gc->cmdTransportInfo.fifoPtr;
  gc->cmdTransportInfo.bumpPos = gc->cmdTransportInfo.fifoPtr + (gc->cmdTransportInfo.bumpSize);
  if (gc->cmdTransportInfo.bumpPos > gc->cmdTransportInfo.fifoEnd)
    gc->cmdTransportInfo.bumpPos = gc->cmdTransportInfo.fifoEnd;

  GR_ASSERT(gc->cmdTransportInfo.bumpPos != 0);
  GR_ASSERT(gc->cmdTransportInfo.lastBump != 0);

#undef FN_NAME
} /* _grCheckForBump */

struct cmdTransportInfo* FX_CALL
_grCommandTransportInfo(void)
{
#define FN_NAME "_grCommandTransportInfo"
  GR_BEGIN_NOFIFOCHECK_NORET(FN_NAME"()\n", 98);

  /* validate the state so that the packet headers are correct */
  if (gc->state.invalid) _grValidateState();

  GR_RETURN(&gc->cmdTransportInfo);
#undef FN_NAME
} /* _grCommandTransportInfo */

void 
_FifoFlush( void ) 
{
#define FN_NAME "_FifoFlush"
  GR_DCL_GC;
  
  _grCommandTransportMakeRoom(0, __FILE__, __LINE__);
#undef FN_NAME
} /* _FifoFlush */

AnyPtr _grHwFifoPtrSlave(FxU32 slave, FxBool ignored);

void FX_CALL
_grCommandTransportMakeRoom(const FxI32 blockSize, const char* fName, const int fLine)
{
#define FN_NAME "_FifoMakeRoom"

  GR_BEGIN_NOFIFOCHECK(FN_NAME, 400);

  if ( !gc->windowed ) {
    /* Check here to see if we have a valid context since the last time
     * we checked. This is to protect us from loosing our context before
     * we wrap check the current hw fifo pointer which is going to be the
     * 2d driver's fifo if we lost our context.
     */
    gc->contextP = 1; /* always has context in DRI */
    if (gc->contextP) {
      AnyPtr wrapAddr = 0x00UL;
      FxU32 checks;

      GR_ASSERT(blockSize > 0);
      GR_ASSERT((FxU32)blockSize < gc->cmdTransportInfo.fifoSize);
      FIFO_ASSERT();
      
      /* Nasty performance testing hack */
      if(_GlideRoot.environment.noHW) {
        /* Cribbed from grSstWinOpen() */
        struct cmdTransportInfo *gcFifo = &gc->cmdTransportInfo;        
        gcFifo->roomToEnd = gcFifo->fifoSize - FIFO_END_ADJUST;
        gcFifo->fifoRoom  = gcFifo->roomToReadPtr = gcFifo->roomToEnd - sizeof( FxU32 );
    
        /* Set initial fifo state. hw read and sw write pointers at
        * start of the fifo.
        */
        gcFifo->fifoPtr  = gcFifo->fifoStart;
        gcFifo->fifoRead = HW_FIFO_PTR( FXTRUE );
        return;
      }
        
    /* Update the roomToXXX values w/ the # of writes since the last
     * fifo stall/wrap.  
     */
      {
        const FxI32 writes = (MIN(gc->cmdTransportInfo.roomToReadPtr, gc->cmdTransportInfo.roomToEnd) -
                              gc->cmdTransportInfo.fifoRoom);
        
        gc->cmdTransportInfo.roomToReadPtr   -= writes;
        gc->cmdTransportInfo.roomToEnd       -= writes;
        
#if GDBG_INFO_ON
        GDBG_INFO_MORE(gc->myLevel, ": (%s : %d)\n"
                       "\tfifoBlock: (0x%X : 0x%X)\n"
                       "\tfifoRoom: (0x%X : 0x%X) : (0x%X : 0x%X)\n"
                       "\tfifo hw: (0x%X : 0x%X)\n",
                       ((fName == NULL) ? "Unknown" : fName), fLine,
                       (((AnyPtr)gc->cmdTransportInfo.fifoPtr - (AnyPtr)gc->cmdTransportInfo.fifoStart) + 
                        (FxU32)gc->cmdTransportInfo.fifoOffset),
                       blockSize,
                       gc->cmdTransportInfo.roomToReadPtr, gc->cmdTransportInfo.roomToEnd, 
                       gc->cmdTransportInfo.fifoRoom, writes,
                       HW_FIFO_PTR(FXTRUE) - (AnyPtr)gc->rawLfb, gc->cmdTransportInfo.fifoRead);
        
#endif /* GDBG_INFO_ON */
        
        ASSERT_FAULT_IMMED((gc->cmdTransportInfo.roomToReadPtr >= 0) && 
                           (gc->cmdTransportInfo.roomToEnd >= 0));
      }
      
      
      /* Bump & Grind if called for */
      if (!gc->cmdTransportInfo.autoBump)
        GR_BUMP_N_GRIND;
      
      checks = 0;
      
  again:
      /* do we need to stall? */
      {
        AnyPtr lastHwRead = gc->cmdTransportInfo.fifoRead;
        FxI32 roomToReadPtr = gc->cmdTransportInfo.roomToReadPtr;
        
        while (roomToReadPtr < blockSize) {
          AnyPtr curReadPtr = HW_FIFO_PTR(FXTRUE);
          FxI32 curReadDist = curReadPtr - lastHwRead;

          /* Handle slave chips.  This code lifted from cvg and modified
           * to deal with multiple slave chips. */
          if(gc->chipCount > 1) {
            FxU32 slave;
            for(slave = 1; slave < gc->chipCount; slave++) {              
              const AnyPtr slaveReadPtr = _grHwFifoPtrSlave(slave, 0);
              const FxU32 slaveReadDist = (slaveReadPtr - lastHwRead);
              FxI32 distSlave = (FxI32)slaveReadDist;
              FxI32 distMaster = (FxI32)curReadDist;

              GR_ASSERT((slaveReadPtr >= (AnyPtr)gc->cmdTransportInfo.fifoStart) &&
                        (slaveReadPtr < (AnyPtr)gc->cmdTransportInfo.fifoEnd));
          
              /* Get the actual absolute distance to the respective fifo ptrs */
              if (distSlave < 0) distSlave += (FxI32)gc->cmdTransportInfo.fifoSize - FIFO_END_ADJUST;
              if (distMaster < 0) distMaster += (FxI32)gc->cmdTransportInfo.fifoSize - FIFO_END_ADJUST;

              /* Is this slave closer than the master? */
              if (distSlave < distMaster) {
                
                curReadDist = slaveReadDist;
                curReadPtr = slaveReadPtr;
              }
            }
          }

          checks++;
          
#ifdef GLIDE_DEBUG
          if (checks > 1000) {
            {
            const FxU32
              baseAddrL = GR_CAGP_GET(baseAddrL),
              baseSize = GR_CAGP_GET(baseSize),
              readPtrL = GR_CAGP_GET(readPtrL),
              aMin = GR_CAGP_GET(aMin),
              aMax = GR_CAGP_GET(aMax),
              depth = GR_CAGP_GET(depth),
              holeCount = GR_CAGP_GET(holeCount);
            
            GDBG_INFO(80,"Fifo check timeout:\n");
            GDBG_INFO(80,"\tbaseAddrL = 0x%x\n", baseAddrL);
            GDBG_INFO(80,"\tbaseSize  = 0x%x\n", baseSize);
            GDBG_INFO(80,"\treadPtrL  = 0x%x\n", readPtrL);
            GDBG_INFO(80,"\tdepth     = 0x%x\n", depth);
            GDBG_INFO(80,"\tholeCount = 0x%x\n", holeCount);
            GDBG_INFO(80,"\taMin      = 0x%x\n", aMin);
            GDBG_INFO(80,"\taMax      = 0x%x\n", aMax);
                
            if ((readPtrL < (baseAddrL << 12)) ||
                (readPtrL > ((baseAddrL + baseSize + 1) << 12))) {
              GDBG_PRINTF("FATAL ERROR:  Read Pointer out of command buffer extents\n");
              exit(-1);
            }
            }

            if (gc->chipCount > 1)
            {
              FxU32 chip ;

              for (chip = 0 ;
                   chip < gc->chipCount - 1 ;
                   chip++)
              {
                const FxU32
                  baseAddrL = GR_SLAVE_CAGP_GET(chip, baseAddrL),
                  baseSize = GR_SLAVE_CAGP_GET(chip, baseSize),
                  readPtrL = GR_SLAVE_CAGP_GET(chip, readPtrL),
                  aMin = GR_SLAVE_CAGP_GET(chip, aMin),
                  aMax = GR_SLAVE_CAGP_GET(chip, aMax),
                  depth = GR_SLAVE_CAGP_GET(chip, depth),
                  holeCount = GR_SLAVE_CAGP_GET(chip, holeCount);
                
                GDBG_INFO(80,"Fifo check timeout slave 0x%x:\n", chip);
                GDBG_INFO(80,"\tbaseAddrL = 0x%x\n", baseAddrL);
                GDBG_INFO(80,"\tbaseSize  = 0x%x\n", baseSize);
                GDBG_INFO(80,"\treadPtrL  = 0x%x\n", readPtrL);
                GDBG_INFO(80,"\tdepth     = 0x%x\n", depth);
                GDBG_INFO(80,"\tholeCount = 0x%x\n", holeCount);
                GDBG_INFO(80,"\taMin      = 0x%x\n", aMin);
                GDBG_INFO(80,"\taMax      = 0x%x\n", aMax);
              }
            }

            checks = 0;
          }
#endif /* GLIDE_DEBUG */
          GR_ASSERT((curReadPtr >= (AnyPtr)gc->cmdTransportInfo.fifoStart) &&
                    (curReadPtr < (AnyPtr)gc->cmdTransportInfo.fifoEnd));
              
          roomToReadPtr += curReadDist;
              
          gc->stats.fifoStalls++;
          gc->stats.fifoStallDepth += GR_CAGP_GET(depth);
              
          /* Have we wrapped yet? */
          if (lastHwRead > curReadPtr) roomToReadPtr += (FxI32)gc->cmdTransportInfo.fifoSize - FIFO_END_ADJUST;
          lastHwRead = curReadPtr;
        }
          
        GR_ASSERT((lastHwRead >= (AnyPtr)gc->cmdTransportInfo.fifoStart) &&
                  (lastHwRead < (AnyPtr)gc->cmdTransportInfo.fifoEnd));
          
        /* Update cached copies */
        gc->cmdTransportInfo.fifoRead = lastHwRead;
        gc->cmdTransportInfo.roomToReadPtr = roomToReadPtr;
          
        GDBG_INFO(gc->myLevel + 10, 
                  "  Wait: (0x%X : 0x%X) : 0x%X\n", 
                  gc->cmdTransportInfo.roomToReadPtr, gc->cmdTransportInfo.roomToEnd,
                  gc->cmdTransportInfo.fifoRead);
      }
      
      /* Do we need to wrap to front? */
      if (gc->cmdTransportInfo.roomToEnd <= blockSize) {
        GDBG_INFO(gc->myLevel + 10, 
                  "  Pre-Wrap: (0x%X : 0x%X) : 0x%X\n", 
                  gc->cmdTransportInfo.roomToReadPtr, gc->cmdTransportInfo.roomToEnd,
                  gc->cmdTransportInfo.fifoRead);
          
        /* Set the jsr packet. 
         * NB: This command must be fenced.
         */
        FIFO_ASSERT();
        {
          P6FENCE;
          if (!gc->cmdTransportInfo.autoBump) {
#if __POWERPC__ && PCI_BUMP_N_GRIND           
            SET_FIFO(*gc->cmdTransportInfo.fifoPtr++, gc->cmdTransportInfo.fifoJmpHdr[0]);
            FIFO_CACHE_FLUSH(gc->cmdTransportInfo.fifoPtr);
            GR_CAGP_SET(bump, 1);
#else /* !__POWERPC__ && !PCI_BUMP_N_GRIND */             
            SET(*gc->cmdTransportInfo.fifoPtr++, gc->cmdTransportInfo.fifoJmpHdr[0]);
            SET(*gc->cmdTransportInfo.fifoPtr++, gc->cmdTransportInfo.fifoJmpHdr[1]);
            GR_CAGP_SET(bump, 2);
#endif /* !__POWERPC__ && !PCI_BUMP_N_GRIND */
            gc->cmdTransportInfo.lastBump = gc->cmdTransportInfo.fifoStart;
          } else {
            SET(*gc->cmdTransportInfo.fifoPtr, gc->cmdTransportInfo.fifoJmpHdr[0]);
          }
        }
          
        P6FENCE;
    
        wrapAddr = (AnyPtr)gc->cmdTransportInfo.fifoPtr;
          
        /* Update roomXXX fields for the actual wrap */
        gc->cmdTransportInfo.roomToReadPtr -= gc->cmdTransportInfo.roomToEnd;
        gc->cmdTransportInfo.roomToEnd = gc->cmdTransportInfo.fifoSize - FIFO_END_ADJUST;
          
#if GLIDE_USE_DEBUG_FIFO
        gc->stats.fifoWraps++;
        gc->stats.fifoWrapDepth += GR_GET(hw->cmdFifoDepth);
#endif /* GLIDE_USE_DEBUG_FIFO */
          
        /* Reset fifo ptr to start */ 
        gc->cmdTransportInfo.fifoPtr = gc->cmdTransportInfo.fifoStart;
          
        /*  We havn't really fenced here, but we set the lastFence for */
        /*  later calculations  */  
        gc->cmdTransportInfo.lastFence = gc->cmdTransportInfo.fifoStart;

        GDBG_INFO(gc->myLevel + 10, 
                  "  Post-Wrap: (0x%X : 0x%X) : 0x%X\n", 
                  gc->cmdTransportInfo.roomToReadPtr, gc->cmdTransportInfo.roomToEnd,
                  gc->cmdTransportInfo.fifoRead);
          
        goto again;
      }
      
      /* compute room left */
      gc->cmdTransportInfo.fifoRoom = MIN(gc->cmdTransportInfo.roomToReadPtr, gc->cmdTransportInfo.roomToEnd);
      
      GDBG_INFO(gc->myLevel, FN_NAME"_Done:\n"
                "\tfifoBlock: (0x%X : 0x%X)\n"
                "\tfifoRoom: (0x%X : 0x%X : 0x%X)\n"
                "\tfifo hw: (0x%X : 0x%X) : (0x%X : 0x%X : 0x%X)\n",
                (((AnyPtr)gc->cmdTransportInfo.fifoPtr - (AnyPtr)gc->cmdTransportInfo.fifoStart) + 
                 (FxU32)gc->cmdTransportInfo.fifoOffset),
                blockSize,
                gc->cmdTransportInfo.roomToReadPtr, 
                gc->cmdTransportInfo.roomToEnd, gc->cmdTransportInfo.fifoRoom,
                HW_FIFO_PTR(FXTRUE) - (AnyPtr)gc->rawLfb, gc->cmdTransportInfo.fifoRead, 
                GR_CAGP_GET(depth), GR_CAGP_GET(holeCount), GR_GET(hw->status));
      
      FIFO_ASSERT();
    }
    else {
      /*
      ** reset the function pointers when the context is gone
      */ 
      GrTriSetupProcArchVector* 
        curTriProcs = _GlideRoot.deviceArchProcs.curTriProcs;
      GrVertexListProc*
        curVertexListProcs = _GlideRoot.deviceArchProcs.curVertexListProcs;
      
      _GlideRoot.deviceArchProcs.curTriProcs        = _GlideRoot.deviceArchProcs.nullTriProcs;
      _GlideRoot.deviceArchProcs.curVertexListProcs = _GlideRoot.deviceArchProcs.nullVertexListProcs;
      
      _GlideRoot.deviceArchProcs.nullTriProcs        = curTriProcs;
      _GlideRoot.deviceArchProcs.nullVertexListProcs = curVertexListProcs;
      
      gc->archDispatchProcs.texDownloadProcs  = _GlideRoot.deviceArchProcs.nullTexProcs;
      gc->archDispatchProcs.drawTrianglesProc = _GlideRoot.deviceArchProcs.nullDrawTrisProc;
      gc->archDispatchProcs.coorModeTriVector = (*_GlideRoot.deviceArchProcs.curTriProcs) + GR_WINDOW_COORDS;
      gc->archDispatchProcs.drawVertexList    = _GlideRoot.deviceArchProcs.curVertexListProcs[GR_WINDOW_COORDS];
      gc->triSetupProc = CUR_TRI_PROC(FXFALSE, (gc->state.cull_mode != GR_CULL_DISABLE));
    }
  }
  GR_TRACE_EXIT(FN_NAME);
#undef FN_NAME
}

void
_grH3FifoDump_Linear(const FxU32* const linearPacketAddr)
{
#ifdef GDBG_INFO_ON
  FXUNUSED(h3SstIORegNames);
#endif
}


AnyPtr
_grHwFifoPtr(FxBool ignored)
{
  AnyPtr rVal = 0;

  FxU32 status, readPtrL1, readPtrL2;
  FxU32 chip ; /* AJB SLI MAYHEM */
  GR_DCL_GC;

  FXUNUSED(ignored);

  if ( gc->windowed ) {
    rVal = 0;
  } else {
    if (0 && gc->chipCount > 1)
    {
      /*
      do {
        readPtrL1 = GET(gc->cRegs->cmdFifo0.readPtrL);
        for (chip = 0 ;
             chip < gc->chipCount - 1 ;
             chip++)
        readPtrL1 = ((readPtrL1 > gc->slaveCRegs[chip]->cmdFifo0.readPtrL)
                     ? readPtrL1 : gc->slaveCRegs[chip]->cmdFifo0.readPtrL) ;
#if __POWERPC__      
        status = GET(gc->ioRegs->status);
#else      
        status = _grSstStatus();
#endif      
        readPtrL2 = GET(gc->cRegs->cmdFifo0.readPtrL);
        for (chip = 0 ;
             chip < gc->chipCount - 1 ;
             chip++)
        readPtrL2 = ((readPtrL2 < gc->slaveCRegs[chip]->cmdFifo0.readPtrL)
                     ? readPtrL2 : gc->slaveCRegs[chip]->cmdFifo0.readPtrL) ;
      } while (readPtrL1 != readPtrL2);
      */

        readPtrL2 = GET(gc->cRegs->cmdFifo0.readPtrL);
#if __POWERPC__      
        status = GET(gc->ioRegs->status);
#else      
        status = _grSstStatus();
#endif      
        for (chip = 0 ;
             chip < gc->chipCount - 1 ;
             chip++)
        readPtrL2 = ((readPtrL2 < gc->slaveCRegs[chip]->cmdFifo0.readPtrL)
                     ? readPtrL2 : gc->slaveCRegs[chip]->cmdFifo0.readPtrL) ;
    }
    else
    {
      do {
        readPtrL1 = GET(gc->cRegs->cmdFifo0.readPtrL);
#if __POWERPC__      
        status = GET(gc->ioRegs->status);
#else      
        status = _grSstStatus();
#endif      
        readPtrL2 = GET(gc->cRegs->cmdFifo0.readPtrL);
      } while (readPtrL1 != readPtrL2);
    }
    rVal = (((AnyPtr)gc->cmdTransportInfo.fifoStart) + 
            readPtrL2 - 
            (FxU32)gc->cmdTransportInfo.fifoOffset);
  }
  return rVal;
} /* _grHwFifoPtr */


AnyPtr
_grHwFifoPtrSlave(FxU32 slave, FxBool ignored)
{
  AnyPtr rVal = 0;

  FxU32 status, readPtrL1, readPtrL2;
  GR_DCL_GC;

  FXUNUSED(ignored);

  do {
    readPtrL1 = GET(gc->slaveCRegs[slave-1]->cmdFifo0.readPtrL);

    status = GET(gc->slaveSstRegs[slave-1]->status);

    readPtrL2 = GET(gc->slaveCRegs[slave-1]->cmdFifo0.readPtrL);
  } while (readPtrL1 != readPtrL2);

  rVal = (((AnyPtr)gc->cmdTransportInfo.fifoStart) + 
          readPtrL2 - 
          (FxU32)gc->cmdTransportInfo.fifoOffset);
  
  return rVal;
} /* _grHwFifoPtr */


#if defined( FIFO_ASSERT_FULL ) 

void 
_fifoAssertFull( void ) 
{
  GR_DCL_GC;
  
  if ( !gc->windowed ) {
    if ((gFifoCheckCount++ & kFifoCheckMask) == 0) {
      const FxU32 cmdFifoDepth = GR_GET(((SstRegs*)(gc->reg_ptr))->cmdFifoDepth);
      const FxU32 maxFifoDepth = ((gc->cmdTransportInfo.fifoSize - FIFO_END_ADJUST) >> 2);
      if(cmdFifoDepth > maxFifoDepth) { 
        gdbg_printf("cmdFifoDepth > size: 0x%X : 0x%Xn", 
                    cmdFifoDepth, maxFifoDepth); 
        ASSERT_FAULT_IMMED(cmdFifoDepth <= maxFifoDepth); 
      } else if (cmdFifoDepth + (gc->cmdTransportInfo.fifoRoom >> 2) > maxFifoDepth) { 
        gdbg_printf("cmdFifoDepth + fifoRoom > size: (0x%X : 0x%X) : 0x%Xn", 
                    cmdFifoDepth, (gc->cmdTransportInfo.fifoRoom >> 2), maxFifoDepth); 
        ASSERT_FAULT_IMMED(cmdFifoDepth + (gc->cmdTransportInfo.fifoRoom >> 2) <= maxFifoDepth); 
      } 
    } 
    ASSERT_FAULT_IMMED(HW_FIFO_PTR(FXTRUE) >= (FxU32)gc->cmdTransportInfo.fifoStart); 
    ASSERT_FAULT_IMMED(HW_FIFO_PTR(FXTRUE) < (FxU32)gc->cmdTransportInfo.fifoEnd); 
    ASSERT_FAULT_IMMED((FxU32)gc->cmdTransportInfo.fifoRoom < gc->cmdTransportInfo.fifoSize); 
    ASSERT_FAULT_IMMED((FxU32)gc->cmdTransportInfo.fifoPtr < (FxU32)gc->cmdTransportInfo.fifoEnd);
  }
}
        
#endif

#if GLIDE_DEBUG
void 
_reg_group_begin_internal( FxU32 __chipId, FxU32 __regBase, FxU32 __groupNum, 
                           FxU32 __groupMask, FxU32 __pktHdr, FxU32 __checkP, 
                           volatile FxU32 *_regGroupFifoPtr ) 
{
  GR_DCL_GC;

  if ( !gc->windowed ) {
    GR_ASSERT(((__pktHdr) & 0xE0000000UL) == 0x00UL); 
    FIFO_ASSERT(); 
    GR_ASSERT(GET(gc->cRegs->cmdFifo0.readPtrL) >= gc->cmdTransportInfo.fifoOffset);
    GR_ASSERT(GET(gc->cRegs->cmdFifo0.readPtrL) < (gc->cmdTransportInfo.fifoOffset + gc->cmdTransportInfo.fifoSize));

    GDBG_INFO(120, "REG_GROUP_BEGIN:\n");
    GDBG_INFO(120, "\tFile: %s Line %d\n", __FILE__, __LINE__);
    GDBG_INFO(120, "\t_regGroupFifoPtr:  0x%x\n", (FxU32)_regGroupFifoPtr - (FxU32)gc->rawLfb);
    GDBG_INFO(120, "\t__pktHdr:          0x%x\n", __pktHdr);
    GDBG_INFO(120, "\t__groupNum:        0x%x\n", __groupNum);
    GDBG_INFO(120, "\t__groupMask:       0x%x\n", (__groupMask));
    GDBG_INFO(120, "\t__chipId:          0x%x\n", __chipId);
    GDBG_INFO(120, "\t__regBase:         0x%x\n", __regBase);
    GDBG_INFO(120, "\tfifoPtr:           0x%x\n", (FxU32)gc->cmdTransportInfo.fifoPtr - (FxU32)gc->cmdTransportInfo.fifoStart - (FxU32)gc->rawLfb); 
    GDBG_INFO(120, "\tfifoRoom:          0x%x\n", gc->cmdTransportInfo.fifoRoom);
    GDBG_INFO(120, "\treadPtrL:          0x%x\n", GET(gc->cRegs->cmdFifo0.readPtrL));
  }
}

void
_reg_group_begin_internal_wax( FxU32 __regBase,
                               FxU32 __groupNum,    
                               FxU32 __groupMask,            
                               FxU32 __pktHdr,               
                               FxU32 __checkP,               
                               volatile FxU32 *__regGroupFifoPtr ) 
{
  GR_DCL_GC;
    
  if ( !gc->windowed ) {
    GR_ASSERT(((__pktHdr) & 0xE0000000UL) == 0x00UL); 
    FIFO_ASSERT(); 
    GR_ASSERT(GET(gc->cRegs->cmdFifo0.readPtrL) >= gc->cmdTransportInfo.fifoOffset);
    GR_ASSERT(GET(gc->cRegs->cmdFifo0.readPtrL) < (gc->cmdTransportInfo.fifoOffset + gc->cmdTransportInfo.fifoSize));

    GDBG_INFO(220, "REG_GROUP_BEGIN_WAX:\n");
    GDBG_INFO(220, "\tFile: %s Line %d\n", __FILE__, __LINE__);
    GDBG_INFO(220, "\t_regGroupFifoPtr:  0x%x\n", 
              (FxU32)__regGroupFifoPtr - (FxU32)gc->rawLfb);
    GDBG_INFO(220, "\t__pktHdr:           0x%x\n", __pktHdr);
    GDBG_INFO(220, "\t__groupNum:         0x%x\n", __groupNum);
    GDBG_INFO(220, "\t__groupMask:        0x%x\n", (__groupMask));
    GDBG_INFO(220, "\t__regBase:          0x%x\n", __regBase);
    GDBG_INFO(220, "\tfifoPtr:            0x%x\n", (FxU32)gc->cmdTransportInfo.fifoPtr - (FxU32)gc->cmdTransportInfo.fifoStart - (FxU32)gc->rawLfb); 
    GDBG_INFO(220, "\tfifoRoom:           0x%x\n", gc->cmdTransportInfo.fifoRoom);
    GDBG_INFO(220, "\treadPtrL:           0x%x\n", GET(gc->cRegs->cmdFifo0.readPtrL));
    GDBG_INFO(220, "\tStart Reg:          0x%x\n", (__pktHdr & 0x7fff) >> 3);
    GDBG_INFO(220, "\tReg Mask:           0x%x\n", (__pktHdr >> 15) & 0x3fff);
    GDBG_INFO(220, "\tReg Type:           %s\n", ((__pktHdr >> 14) & 1) ? "2D" : "3D");
  }
    
}
#endif /* GLIDE_DEBUG */
#endif /* USE_PACKET_FIFO */

#ifdef __alpha__
unsigned char _fxget8( unsigned char *pval ) {
    __asm__ __volatile__("mb": : :"memory");
    return( *pval );
}
unsigned short _fxget16( unsigned short *pval ) {
    __asm__ __volatile__("mb": : :"memory");
    return( *pval );
}
unsigned int _fxget32( unsigned int *pval ) {
    __asm__ __volatile__("mb": : :"memory");
    return( *pval );
}
#endif /* __alpha__ */

void
_grImportFifo(AnyPtr fifoPtr, AnyPtr fifoRead) {
  struct cmdTransportInfo* gcFifo;
  FxU32 readPos;
  GR_DCL_GC;

#if 1
  int dummy, d;

  do {
    dummy=GET(gc->cRegs->cmdFifo0.depth);
    d=GET(gc->cRegs->cmdFifo0.depth);
  } while (dummy || d);
  do {
    dummy = GET(gc->cRegs->cmdFifo0.readPtrL);
    readPos = GET(gc->cRegs->cmdFifo0.readPtrL);
  } while (dummy!=readPos);
  gcFifo=&gc->cmdTransportInfo;
  readPos=readPos-gcFifo->fifoOffset;
  gcFifo->fifoPtr = gcFifo->fifoStart + (readPos>>2);
  gcFifo->fifoRead = (AnyPtr)gcFifo->fifoPtr;
#else
  gcFifo=&gc->cmdTransportInfo;
  gcFifo->fifoPtr = gc->rawLfb+(fifoPtr>>2);
  gcFifo->fifoRead = ((int)gc->rawLfb)+fifoRead;
#endif
  gcFifo->roomToReadPtr = gcFifo->fifoRead-((AnyPtr)gcFifo->fifoPtr)-FIFO_END_ADJUST-sizeof(FxU32);
  if (gcFifo->roomToReadPtr<0) gcFifo->roomToReadPtr+=gcFifo->fifoSize;
  gcFifo->roomToEnd = gcFifo->fifoSize - 
    ((gcFifo->fifoPtr-gcFifo->fifoStart)<<2) -
    FIFO_END_ADJUST;
  gcFifo->fifoRoom = MIN(gcFifo->roomToEnd, gcFifo->roomToReadPtr);
  if (!gc->cmdTransportInfo.autoBump) {
    gcFifo->lastBump = gcFifo->fifoPtr;
    gcFifo->bumpPos = gcFifo->fifoPtr + gcFifo->bumpSize;
  }
}

void
_grExportFifo(FxU32 *fifoPtr, FxU32 *fifoRead) {
  struct cmdTransportInfo* gcFifo;
  GR_DCL_GC;
  gcFifo=&gc->cmdTransportInfo;
  *fifoPtr=(gcFifo->fifoPtr-gc->rawLfb)<<2;
  *fifoRead=(gcFifo->fifoRead-(AnyPtr)gc->rawLfb);
}

int
grFifoGetStalls() {
  GR_DCL_GC;
  return gc->stats.fifoStalls;
}
