/*
 ** THIS SOFTWARE IS SUBJECT TO COPYRIGHT PROTECTION AND IS OFFERED ONLY
 ** PURSUANT TO THE 3DFX GLIDE GENERAL PUBLIC LICENSE. THERE IS NO RIGHT
 ** TO USE THE GLIDE TRADEMARK WITHOUT PRIOR WRITTEN PERMISSION OF 3DFX
 ** INTERACTIVE, INC. A COPY OF THIS LICENSE MAY BE OBTAINED FROM THE 
 ** DISTRIBUTOR OR BY CONTACTING 3DFX INTERACTIVE INC(info@3dfx.com). 
 ** THIS PROGRAM IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 ** EXPRESSED OR IMPLIED. SEE THE 3DFX GLIDE GENERAL PUBLIC LICENSE FOR A
 ** FULL TEXT OF THE NON-WARRANTY PROVISIONS.  
 ** 
 ** USE, DUPLICATION OR DISCLOSURE BY THE GOVERNMENT IS SUBJECT TO
 ** RESTRICTIONS AS SET FORTH IN SUBDIVISION (C)(1)(II) OF THE RIGHTS IN
 ** TECHNICAL DATA AND COMPUTER SOFTWARE CLAUSE AT DFARS 252.227-7013,
 ** AND/OR IN SIMILAR OR SUCCESSOR CLAUSES IN THE FAR, DOD OR NASA FAR
 ** SUPPLEMENT. UNPUBLISHED RIGHTS RESERVED UNDER THE COPYRIGHT LAWS OF
 ** THE UNITED STATES.  
 ** 
 ** COPYRIGHT 3DFX INTERACTIVE, INC. 1999, ALL RIGHTS RESERVED
 **
 ** $Header: /home/cvsroot/driglide/driglide/h5/glide3/src/gxdraw.c,v 1.2 2003/01/15 04:27:57 anholt Exp $
 **
 */

#include <3dfx.h>
#include <glidesys.h>

#define FX_DLL_DEFINITION
#include <fxdll.h>
#include <glide.h>
#include "fxglide.h"
#include "fxcmd.h"

#ifdef GDBG_INFO_ON
/* Some debugging information */
static char *indexNames[] = {  
  "GR_VERTEX_X_OFFSET",         /* 0 */
  "GR_VERTEX_Y_OFFSET",         /* 1 */
  "GR_VERTEX_Z_OFFSET",         /* 2 */
  "GR_VERTEX_R_OFFSET",         /* 3 */
  "GR_VERTEX_G_OFFSET",         /* 4 */
  "GR_VERTEX_B_OFFSET",         /* 5 */
  "GR_VERTEX_OOZ_OFFSET",       /* 6 */
  "GR_VERTEX_A_OFFSET",         /* 7 */
  "GR_VERTEX_OOW_OFFSET",       /* 8 */
  "GR_VERTEX_SOW_TMU0_OFFSET",  /* 9 */
  "GR_VERTEX_TOW_TMU0_OFFSET",  /* 10 */
  "GR_VERTEX_OOW_TMU0_OFFSET",  /* 11 */
  "GR_VERTEX_SOW_TMU1_OFFSET",  /* 12 */
  "GR_VERTEX_TOW_TMU1_OFFSET",  /* 13 */
  "GR_VERTEX_OOW_TMU1_OFFSET"   /* 14 */
};  
#endif /* GDBG_INFO_ON */

#if GLIDE_USE_C_TRISETUP

static FxI32
_grTriCull(const void* a, const void* b, const void* c)
{
#define FN_NAME "_grCullTri"
  GR_BEGIN_NOFIFOCHECK_RET(FN_NAME, 85);
  GDBG_INFO_MORE(gc->myLevel, 
                 "(0x%X, 0x%X, 0x%X)\n", 
                 a, b, c);

  {
    const FxI32 
      xindex = (gc->state.vData.vertexInfo.offset >> 2),
      yindex = xindex + 1;
    const float 
      *fa = (const float*)a + xindex,
      *fb = (const float*)b + xindex,
      *fc = (const float*)c + xindex;
    const float 
      dxAB = fa[xindex] - fb[xindex],
      dxBC = fb[xindex] - fc[xindex], 
      dyAB = fa[yindex] - fb[yindex],
      dyBC = fb[yindex] - fc[yindex],
      area = dxAB * dyBC - dxBC * dyAB;
  
    {
      const FxI32 j = *(FxI32*)&area;
      const FxU32 culltest = (gc->state.cull_mode << 31UL);
    
      /* Zero-area triangles are BAD!! */
      if ((j & 0x7FFFFFFF) == 0) {
        GDBG_INFO(291, FN_NAME": Culling (%g %g) (%g %g) (%g %g) : (%g : 0x%X : 0x%X)\n",
                  (fa[0]), (fa[1]), 
                  (fb[0]), (fb[1]), 
                  (fc[0]), (fc[1]), 
                  area, gc->state.cull_mode, culltest);
      
        return 0;
      }
    
      /* Backface culling, use sign bit as test */
      if ((gc->state.cull_mode != GR_CULL_DISABLE) && (((FxI32)(j ^ culltest)) >= 0)) {
        GDBG_INFO(291, FN_NAME": Culling (%g %g) (%g %g) (%g %g) : (%g : 0x%X : 0x%X)\n",
                  (fa[0]), (fa[1]), 
                  (fb[0]), (fb[1]), 
                  (fc[0]), (fc[1]), 
                  area, gc->state.cull_mode, culltest);
      
        return -1;
      }
    }

    return 1;
  }

  GR_END();
#undef FN_NAME
}

static FxI32 
internal_trisetup(const char* FN_NAME, 
                  const FxBool cullP, const FxBool validStateP,
                  const void* a, const void* b, const void* c)
{
  GR_BEGIN_NOFIFOCHECK_RET(FN_NAME, 85);
  GDBG_INFO_MORE(gc->myLevel, 
                 "(0x%X, 0x%X, 0x%X)\n", 
                 a, b, c);

  /* Check to see if the state has to be flushed */
  if (!validStateP) GR_FLUSH_STATE();

  /* Pass the current culling mode? */
  if (cullP) {
    const FxI32
      cullVal = _grTriCull(a, b, c);

    if (cullVal <= 0) return cullVal;
  }

  /* Validate parameter coordinates */
  {
    const FxI32 
      xindex = (gc->state.vData.vertexInfo.offset >> 2),
      yindex = xindex + 1;
    const float 
      *fa = (const float*)a + xindex,
      *fb = (const float*)b + xindex,
      *fc = (const float*)c + xindex;
    const float 
      aX = fa[xindex],
      aY = fa[yindex],
      bX = fb[xindex],
      bY = fb[yindex],
      cX = fc[xindex],
      cY = fc[yindex];

#define kDimThreshX 2048.0f
#define kDimThreshY 2048.0f
    GR_ASSERT((fabs(aX) < kDimThreshX) && (fabs(aY) < kDimThreshY));
    GR_ASSERT((fabs(bX) < kDimThreshX) && (fabs(bY) < kDimThreshY));
    GR_ASSERT((fabs(cX) < kDimThreshX) && (fabs(cY) < kDimThreshY));
#undef kDimThreshX
#undef kDimThreshY
  }

  /* Send triangle parameters */

#if GLIDE_HW_TRI_SETUP && GLIDE_PACKET3_TRI_SETUP
  {
    FxU32
      dataElem,
      i;

    GR_SET_EXPECTED_SIZE(gc->curTriSize, 1);
    TRI_STRIP_BEGIN(kSetupStrip, 3, gc->curVertexSize, SSTCP_PKT3_BDDBDD);
    {
      TRI_SETF(FARRAY(a, 0));
      TRI_SETF(FARRAY(a, sizeof(float)));
      dataElem = 0;
      i = gc->tsuDataList[dataElem];
      while (i != GR_DLIST_END) {
        TRI_SETF(FARRAY(a, i));
        dataElem++;
        i = gc->tsuDataList[dataElem];
      }

      TRI_SETF(FARRAY(b, 0));
      TRI_SETF(FARRAY(b, sizeof(float)));
      dataElem = 0;
      i = gc->tsuDataList[dataElem];
      while (i != GR_DLIST_END) {
        TRI_SETF(FARRAY(b, i));
        dataElem++;
        i = gc->tsuDataList[dataElem];
      }

      TRI_SETF(FARRAY(c, 0));
      TRI_SETF(FARRAY(c, sizeof(float)));
      dataElem = 0;
      i = gc->tsuDataList[dataElem];
      while (i != GR_DLIST_END) {
        TRI_SETF(FARRAY(c, i));
        dataElem++;
        i = gc->tsuDataList[dataElem];
      }
      TRI_END;
      GR_CHECK_SIZE();
    }
  }
#else
  {
    GR_DCL_HW;
    int vectorIndex;
    FxU32 sMode = (gc->cmdTransportInfo.paramMask >> SSTCP_PKT3_PMASK_SHIFT);
    FxU32 paramMask = (sMode & 0xFF);
    FxU32 paramCount;
    const float* vectorArray[3];

    vectorArray[0] = (const float *)a;
    vectorArray[1] = (const float *)b;
    vectorArray[2] = (const float *)c;
    
    /* Convert packet 3 paramMask into sMode format */
    sMode = (paramMask | ((sMode & 0xF000) << 4));
    
    {
      const FxBool hasColor = ((sMode & 0x01) != 0);
      const FxBool hasAlpha = ((sMode & 0x02) != 0);
      const FxBool hasZ = ((sMode & 0x04) != 0);
      const FxBool hasWb = ((sMode & 0x08) != 0);
      const FxBool hasW0 = ((sMode & 0x10) != 0);
      const FxBool hasST0 = ((sMode & 0x20) != 0);
      const FxBool hasW1 = ((sMode & 0x40) != 0);
      const FxBool hasST1 = ((sMode & 0x80) != 0);
      
      /* We always send vertex XY */
      paramCount = 2;
      paramMask = 0x03;
      
      /* Build parameter data for reg group packet */
#if GLIDE_PACKED_RGB
      if (hasColor || hasAlpha) {
        paramCount += 1;
        paramMask |= 0x04;
      }
#else /* !GLIDE_PACKED_RGB */
      if (hasColor) {
        paramCount += 3;
        paramMask |= 0x38;
      }
      if (hasAlpha) {
        paramCount += 1;
        paramMask |= 0x40;
      }
#endif /* !GLIDE_PACKED_RGB */
      
      if (hasZ) {
        paramCount += 1;
        paramMask |= 0x80;
      }
      if (hasWb) {
        paramCount += 1;
        paramMask |= 0x100;
      }
      if (hasW0) {
        paramCount += 1;
        paramMask |= 0x200;
      }
      if (hasST0) {
        paramCount += 2;
        paramMask |= 0xC00;
      }
      if (hasW1) {
        paramCount += 1;
        paramMask |= 0x1000;
      }
      if (hasST1) {
        paramCount += 2;
        paramMask |= 0x6000;
      }
      
      /* Set mode once for teh whole triangle */
      GR_SET_EXPECTED_SIZE(sizeof(FxU32), 1);
      GR_SET(BROADCAST_ID, hw, sSetupMode, sMode);
      GR_CHECK_SIZE();
      
      for(vectorIndex = 0; vectorIndex < sizeof(vectorArray) / sizeof(float*); vectorIndex++) {
        FxU32
          dataElem,
          i;
        const float* a = (const float *)vectorArray[vectorIndex];
        
        REG_GROUP_BEGIN(BROADCAST_ID, sVx, paramCount, paramMask);
        {
          REG_GROUP_SETF(hw, sVx, FARRAY(a, 0));
          REG_GROUP_SETF(hw, sVy, FARRAY(a, sizeof(float)));
          
          dataElem = 0;
          i = gc->tsuDataList[dataElem];

          if (hasColor) {
            REG_GROUP_SETF(hw, sRed, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
            REG_GROUP_SETF(hw, sGreen, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
            REG_GROUP_SETF(hw, sBlue, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
          }
          if (hasAlpha) {
            REG_GROUP_SETF(hw, sAlpha, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
          }
          
          if (hasZ) {
            REG_GROUP_SETF(hw, sVz, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
          }
          if (hasWb) {
            REG_GROUP_SETF(hw, sOowfbi, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
          }
          
          /* TMU0 */
          if (hasW0) {
            REG_GROUP_SETF(hw, sOow0, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
          }
          if (hasST0) {
            REG_GROUP_SETF(hw, sSow0, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
            REG_GROUP_SETF(hw, sTow0, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
          }
          
          /* TMU1 */
          if (hasW1) {
            REG_GROUP_SETF(hw, sOow1, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
          }
          if (hasST1) {
            REG_GROUP_SETF(hw, sSow1, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
            REG_GROUP_SETF(hw, sTow1, FARRAY(a, i));
            dataElem++;
            i = gc->tsuDataList[dataElem];
          }
        }
        REG_GROUP_END();
        
        GR_SET_EXPECTED_SIZE(sizeof(FxU32), 1);
        if (vectorIndex == 0) {
          GR_SET(BROADCAST_ID, hw, sBeginTriCMD, 0);
        } else {
          GR_SET(BROADCAST_ID, hw, sDrawTriCMD, 0);
        }
        GR_CHECK_SIZE();
      }
    }
  }
#endif

  GR_END();

  return FXTRUE;
}

FxI32 FX_CALL 
_trisetup_Default_win_cull_invalid(const void* a, const void* b, const void* c)
{
#define FN_NAME "_trisetup_Default_win_cull_invalid"
  return internal_trisetup(FN_NAME,
                           FXTRUE, FXFALSE,
                           a, b, c);
#undef FN_NAME
}

FxI32 FX_CALL 
_trisetup_Default_win_cull_valid(const void* a, const void* b, const void* c)
{
#define FN_NAME "_trisetup_Default_win_cull_valid"
  return internal_trisetup(FN_NAME,
                           FXTRUE, FXTRUE,
                           a, b, c);
#undef FN_NAME
}

FxI32 FX_CALL 
_trisetup_Default_win_nocull_invalid(const void* a, const void* b, const void* c)
{
#define FN_NAME "_trisetup_Default_win_nocull_invalid"
  return internal_trisetup(FN_NAME,
                           FXFALSE, FXFALSE,
                           a, b, c);
#undef FN_NAME
}

FxI32 FX_CALL 
_trisetup_Default_win_nocull_valid(const void* a, const void* b, const void* c)
{
#define FN_NAME "_trisetup_Default_win_nocull_valid"
  return internal_trisetup(FN_NAME,
                           FXFALSE, FXTRUE,
                           a, b, c);
#undef FN_NAME
}

#endif

FxI32 FX_CALL 
_vptrisetup_cull(const void* a, const void* b, const void* c)
{
#define FN_NAME "_vptrisetup_cull"
  float *vlist[3];

  GR_BEGIN_NOFIFOCHECK_RET(FN_NAME, 85);
  GDBG_INFO_MORE(gc->myLevel, 
                 "(0x%X, 0x%X, 0x%X)\n", 
                 a, b, c);

  vlist[0] = (float *)a; vlist[1] = (float *)b; vlist[2] = (float *)c;
  (*gc->archDispatchProcs.drawTrianglesProc)(GR_VTX_PTR_ARRAY, 3, vlist);
  
  GR_END();

  return 1;
#undef FN_NAME
}
