#include "cpu_detect.h"

int cpu_features=0;
int proclevel=0;

#ifdef __i386__

int x86_level=0;
int has_cpuid=0;

#define GENU	0x756e6547	/* "Genu" */
#define INEI	0x49656e69	/* "ineI" */
#define NTEL	0x6c65746e	/* "ntel" */

#define AUTH	0x68747541	/* "Auth" */
#define ENTI	0x69746e65	/* "enti" */
#define CAMD	0x444d4163	/* "cAMD" */

/* Calls cpuid with an eax of 'ax' and returns the 16 bytes in *p
 * We are compiled with -fPIC, so we can't clobber ebx.
 */
static void
do_cpuid(int ax, int *p)
{
	__asm __volatile("pushl %%ebx\n\t"
			"cpuid\n\t"
			"movl %%ebx, %%esi\n\t"
			"popl %%ebx"
			 : "=a" (p[0]), "=S" (p[1]), "=c" (p[2]), "=d" (p[3])
			 :  "0" (ax));
}

/* From xf86info havecpuid.c 1.11 */
static int
have_cpuid()
{
	unsigned int f1, f2;
	__asm__ volatile("pushfl\n\t"
			"pushfl\n\t"
			"popl %0\n\t"
			"movl %0,%1\n\t"
			"xorl %2,%0\n\t"
			"pushl %0\n\t"
			"popfl\n\t"
			"pushfl\n\t"
			"popl %0\n\t"
			"popfl\n\t"
			: "=&r" (f1), "=&r" (f2)
			: "ir" (0x00200000));
	return ((f1^f2) & 0x00200000) != 0;
}

void cpu_detect() {
	int cpuid0[4];
	int cpuid1[4];
	int cpuidext0[4];
	int maxcpuid;

	if (cpu_features & CPU_FEATURES_INITED)
		return;

	if (!have_cpuid())
		return;
	
	do_cpuid(0, cpuid0);
	do_cpuid(1, cpuid1);
	do_cpuid(0x80000000, cpuidext0);
	maxcpuid = cpuidext0[0];
	
	/* FIXME: Handle larger cpu levels */
	if (((cpuid1[0] >> 8) & 0xf) < 6)
		x86_level = 5;	/* A P5 */
	else
		x86_level = 4; /* Less */
	
	if (cpuid1[3] & (1 << 23)) {
		cpu_features |= CPU_FEATURE_MMX;
	}
	/* FIXME: The SSE features are disabled because we need
	 * to check for OS support.  However, Glide doesn't
	 * use these features. */
	
	if (cpuid1[3] & (1 << 25)) {
		/*cpu_features |= CPU_FEATURE_SSE;*/
	}
	if (cpuid1[3] & (1 << 26)) {
		/*cpu_features |= CPU_FEATURE_SSE2;*/
	}
	if (    cpuid0[1] == AUTH &&
		cpuid0[3] == ENTI &&
		cpuid0[2] == CAMD &&
		maxcpuid >= 0x80000001 )
	{

		int cpuidext1[4];

		do_cpuid(0x80000001, cpuidext1);
		if (cpuidext1[3] & (1 << 31)) {
			cpu_features |= CPU_FEATURE_3DNOW;
		}
		if (cpuidext1[3] & (1 << 30)) {
			cpu_features |= CPU_FEATURE_3DNOWEXT;
		}
		
		
	} else if (
		cpuid0[1] == GENU &&
		cpuid0[3] == INEI &&
		cpuid0[2] == NTEL) {
		/* Nothing extra for intel */
		
	} else {
		/* Unknown processor, no extra features */
		
	}
	cpu_features |= CPU_FEATURES_INITED;
}

int cpu_get_x86level() {
	if (!(cpu_features & CPU_FEATURES_INITED))
		cpu_detect();
	return x86_level;
}

#else /* __i386__ */

void cpu_detect() {
	cpu_features=CPU_FEATURES_INITED;
}

#endif /* __i386__ */
